__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, nmlop
from nml.actions import action0, action1, action2, action2real, action2var, action6, actionD, real_sprite
from nml.ast import general, spriteblock


class Action2Layout(action2.Action2):
    def __init__(self, feature, name, pos, layout, param_registers):
        action2.Action2.__init__(self, feature, name, pos)
        self.layout = layout
        self.param_registers = param_registers

    def resolve_tmp_storage(self):
        for reg in self.param_registers:
            if not self.tmp_locations:
                raise generic.ScriptError(
                    "There are not enough registers available "
                    + "to perform all required computations in switch blocks. "
                    + "Please reduce the complexity of your code.",
                    self.pos,
                )
            location = self.tmp_locations[0]
            self.remove_tmp_location(location, False)
            reg.set_register(location)

    def write(self, file):
        size = self.layout.get_size()
        regs = ["{} : register {:X}".format(reg.name, reg.register) for reg in self.param_registers]
        action2.Action2.write_sprite_start(self, file, size, regs)
        self.layout.write(file)
        file.end_sprite()


class Action2LayoutSpriteType:
    GROUND = 0
    BUILDING = 1
    CHILD = 2


# these keywords are used to identify a ground/building/childsprite
layout_sprite_types = {
    "ground": Action2LayoutSpriteType.GROUND,
    "building": Action2LayoutSpriteType.BUILDING,
    "childsprite": Action2LayoutSpriteType.CHILD,
}


class Action2LayoutSprite:
    def __init__(self, feature, type, layout_registers=None, pos=None, extra_dicts=None):
        self.feature = feature
        self.type = type
        self.layout_registers = layout_registers
        self.pos = pos
        self.extra_dicts = extra_dicts or []
        self.params = {
            "sprite": {"value": None, "validator": self._validate_sprite},
            "recolour_mode": {"value": 0, "validator": self._validate_recolour_mode},
            "palette": {"value": expression.ConstantNumeric(0), "validator": self._validate_palette},
            "always_draw": {"value": 0, "validator": self._validate_always_draw},
            "xoffset": {"value": expression.ConstantNumeric(0), "validator": self._validate_bounding_box},
            "yoffset": {"value": expression.ConstantNumeric(0), "validator": self._validate_bounding_box},
            "zoffset": {"value": expression.ConstantNumeric(0), "validator": self._validate_bounding_box},
            "xextent": {"value": expression.ConstantNumeric(16), "validator": self._validate_bounding_box},
            "yextent": {"value": expression.ConstantNumeric(16), "validator": self._validate_bounding_box},
            "zextent": {"value": expression.ConstantNumeric(16), "validator": self._validate_bounding_box},
            "hide_sprite": {"value": None, "validator": self._validate_hide_sprite},  # Value not used
        }
        for i in self.params:
            self.params[i]["is_set"] = False
            self.params[i]["register"] = None
        self.sprite_from_action1 = False
        self.palette_from_action1 = False
        self.var10_for_sprite = None
        self.var10_for_palette = None

    def is_advanced_sprite(self):
        if self.feature == 0x04:
            return True
        if self.palette_from_action1:
            return True
        return len(self.get_all_registers()) != 0

    def get_registers_size(self):
        # Number of registers to write
        size = len(self.get_all_registers())
        # Add station specific registers
        if self.feature == 0x04:
            if self.var10_for_sprite:
                size += 1
            if self.var10_for_palette:
                size += 1
        # Add 2 for the flags
        size += 2
        return size

    def write_flags(self, file):
        flags = 0
        if self.get_register("hide_sprite") is not None:
            flags |= 1 << 0
        if self.get_register("sprite") is not None:
            flags |= 1 << 1
        if self.get_register("palette") is not None:
            flags |= 1 << 2
        if self.palette_from_action1:
            flags |= 1 << 3
        # for building sprites: bit 4 => xoffset+yoffset, bit 5 => zoffset (x and y always set together)
        # for child sprites: bit 4 => xoffset, bit 5 => yoffset
        if self.type == Action2LayoutSpriteType.BUILDING:
            assert (self.get_register("xoffset") is not None) == (self.get_register("yoffset") is not None)
        if self.get_register("xoffset") is not None:
            flags |= 1 << 4
        nextreg = "zoffset" if self.type == Action2LayoutSpriteType.BUILDING else "yoffset"
        if self.get_register(nextreg) is not None:
            flags |= 1 << 5
        if self.feature == 0x04:
            if self.var10_for_sprite is not None:
                flags |= 1 << 6
            if self.var10_for_palette is not None:
                flags |= 1 << 7
        file.print_wordx(flags)

    def write_register(self, file, name):
        register = self.get_register(name)[0]
        file.print_bytex(register.parameter)

    def write_registers(self, file):
        if self.is_set("hide_sprite"):
            self.write_register(file, "hide_sprite")
        if self.get_register("sprite") is not None:
            self.write_register(file, "sprite")
        if self.get_register("palette") is not None:
            self.write_register(file, "palette")
        if self.get_register("xoffset") is not None:
            self.write_register(file, "xoffset")
        if self.get_register("yoffset") is not None:
            self.write_register(file, "yoffset")
        if self.get_register("zoffset") is not None:
            self.write_register(file, "zoffset")
        if self.feature == 0x04:
            if self.var10_for_sprite is not None:
                file.print_bytex(self.var10_for_sprite)
            if self.var10_for_palette is not None:
                file.print_bytex(self.var10_for_palette)

    def write_sprite_number(self, file):
        num = self.get_sprite_number()
        if isinstance(num, expression.ConstantNumeric):
            num.write(file, 4)
        else:
            file.print_dwordx(0)

    def get_sprite_number(self):
        # Layout of sprite number
        # bit  0 - 13: Sprite number
        # bit 14 - 15: Recolour mode (normal/transparent/remap)
        # bit 16 - 29: Palette sprite number
        # bit 30: Always draw sprite, even in transparent mode
        # bit 31: This is a custom sprite (from action1), not a TTD sprite
        if not self.is_set("sprite"):
            raise generic.ScriptError("'sprite' must be set for this layout sprite", self.pos)

        # Make sure that recolouring is set correctly
        if self.get_param("recolour_mode") == 0 and self.is_set("palette"):
            raise generic.ScriptError("'palette' may not be set when 'recolour_mode' is RECOLOUR_NONE.")
        elif self.get_param("recolour_mode") != 0 and not self.is_set("palette"):
            raise generic.ScriptError("'palette' must be set when 'recolour_mode' is not set to RECOLOUR_NONE.")

        # Add the constant terms first
        sprite_num = self.get_param("recolour_mode") << 14
        if self.get_param("always_draw"):
            sprite_num |= 1 << 30
        if self.sprite_from_action1:
            sprite_num |= 1 << 31

        # Add the sprite
        expr = nmlop.ADD(self.get_param("sprite"), sprite_num, self.pos)
        # Add the palette
        expr = nmlop.ADD(nmlop.SHIFT_LEFT(self.get_param("palette"), 16, self.pos), expr)
        return expr.reduce()

    def get_param(self, name):
        assert name in self.params
        return self.params[name]["value"]

    def is_set(self, name):
        assert name in self.params
        return self.params[name]["is_set"]

    def get_register(self, name):
        assert name in self.params
        return self.params[name]["register"]

    def get_all_registers(self):
        return [self.get_register(name) for name in sorted(self.params) if self.get_register(name) is not None]

    def create_register(self, name, value):
        if (
            # Always copy values from "prepare_layout" into new registers, to prevent "default" from modifying them.
            self.feature != 0x04
            and isinstance(value, expression.StorageOp)
            and value.name == "LOAD_TEMP"
            and isinstance(value.register, expression.ConstantNumeric)
        ):
            store_tmp = None
            load_tmp = action2var.VarAction2Var(0x7D, 0, 0xFFFFFFFF, value.register.value)
        else:
            if self.layout_registers is None:
                store_tmp = action2var.VarAction2StoreTempVar()
            else:
                store_tmp = self.layout_registers.add(value)
            load_tmp = action2var.VarAction2LoadTempVar(store_tmp)
        self.params[name]["register"] = (load_tmp, store_tmp, value)

    def set_param(self, name, value):
        assert isinstance(name, expression.Identifier)
        assert isinstance(value, expression.Expression)
        name = name.value

        if name not in self.params:
            raise generic.ScriptError("Unknown sprite parameter '{}'".format(name), value.pos)
        if self.is_set(name):
            raise generic.ScriptError("Sprite parameter '{}' can be set only once per sprite.".format(name), value.pos)

        self.params[name]["value"] = self.params[name]["validator"](name, value)
        self.params[name]["is_set"] = True

    def _validate_offset(self, offset, spriteset, pos):
        if len(offset) == 0:
            offset = None
        elif len(offset) == 1:
            id_dicts = [
                (spriteset.labels if spriteset else {}, lambda name, val, pos: expression.ConstantNumeric(val, pos))
            ]
            offset = action2var.reduce_varaction2_expr(
                offset[0], action2var.get_scope(self.feature), self.extra_dicts + id_dicts
            )
            if spriteset and isinstance(offset, expression.ConstantNumeric):
                generic.check_range(
                    offset.value,
                    0,
                    len(real_sprite.parse_sprite_data(spriteset)) - 1,
                    "offset within spriteset",
                    pos,
                )
        else:
            raise generic.ScriptError("Expected 0 or 1 parameter, got " + str(len(offset)), pos)
        return offset

    def resolve_spritegroup_ref(self, sg_ref):
        """
        Resolve a reference to a (sprite/palette) sprite group

        @param sg_ref: Reference to a sprite group
        @type sg_ref: L{SpriteGroupRef}

        @return: Sprite number (index of action1 set) to use
        @rtype: L{Expression}
        """
        spriteset = action2.resolve_spritegroup(sg_ref.name)
        offset = self._validate_offset(sg_ref.param_list, spriteset, sg_ref.pos)
        num = action1.get_action1_index(spriteset, self.feature)
        generic.check_range(num, 0, (1 << 14) - 1, "sprite", sg_ref.pos)
        return expression.ConstantNumeric(num), offset

    def _validate_sprite(self, name, value):
        if isinstance(value, expression.SpriteGroupRef):
            assert self.feature != 0x04
            self.sprite_from_action1 = True
            val, offset = self.resolve_spritegroup_ref(value)
            if offset is not None:
                self.create_register(name, offset)
            return val
        elif isinstance(value, StationSpriteset):
            assert self.feature == 0x04
            self.sprite_from_action1 = True
            if value.offset is not None:
                offset = self._validate_offset(value.offset, value.spriteset, value.pos)
                if offset is not None:
                    self.create_register(name, offset)
            self.var10_for_sprite = value.var10
            return expression.ConstantNumeric(0x42D)
        else:
            self.sprite_from_action1 = False
            if isinstance(value, expression.ConstantNumeric):
                generic.check_range(value.value, 0, (1 << 14) - 1, "sprite", value.pos)
                return value
            if value.supported_by_actionD(raise_error=False):
                return value
            self.create_register(name, value)
            return expression.ConstantNumeric(0)

    def _validate_recolour_mode(self, name, value):
        if not isinstance(value, expression.ConstantNumeric):
            raise generic.ScriptError("Expected a compile-time constant.", value.pos)

        if value.value not in (0, 1, 2):
            raise generic.ScriptError(
                "Value of 'recolour_mode' must be RECOLOUR_NONE, RECOLOUR_TRANSPARENT or RECOLOUR_REMAP."
            )
        return value.value

    def _validate_palette(self, name, value):
        if isinstance(value, expression.SpriteGroupRef):
            assert self.feature != 0x04
            self.palette_from_action1 = True
            val, offset = self.resolve_spritegroup_ref(value)
            if offset is not None:
                self.create_register(name, offset)
            return val
        elif isinstance(value, StationSpriteset):
            assert self.feature == 0x04
            self.palette_from_action1 = True
            if value.offset is not None:
                offset = self._validate_offset(value.offset, value.spriteset, value.pos)
                if offset is not None:
                    self.create_register(name, offset)
            self.var10_for_palette = value.var10
            return expression.ConstantNumeric(0x42D)
        else:
            self.palette_from_action1 = False
            if isinstance(value, expression.ConstantNumeric):
                generic.check_range(value.value, 0, (1 << 14) - 1, "palette", value.pos)
                return value
            if value.supported_by_actionD(raise_error=False):
                return value
            self.create_register(name, value)
            return expression.ConstantNumeric(0)

    def _validate_always_draw(self, name, value):
        if not isinstance(value, expression.ConstantNumeric):
            raise generic.ScriptError("Expected a compile-time constant number.", value.pos)
        # Not valid for ground sprites, raise error
        if self.type == Action2LayoutSpriteType.GROUND:
            raise generic.ScriptError(
                "'always_draw' may not be set for groundsprites, these are always drawn anyways.", value.pos
            )

        if value.value not in (0, 1):
            raise generic.ScriptError("Value of 'always_draw' should be 0 or 1", value.pos)
        return value.value

    def _validate_bounding_box(self, name, value):
        if self.type == Action2LayoutSpriteType.GROUND:
            raise generic.ScriptError(name + " can not be set for ground sprites", value.pos)
        elif self.type == Action2LayoutSpriteType.CHILD:
            if name not in ("xoffset", "yoffset"):
                raise generic.ScriptError(name + " can not be set for child sprites", value.pos)
            if isinstance(value, expression.ConstantNumeric):
                generic.check_range(value.value, 0, 255, name, value.pos)
                return value
        else:
            assert self.type == Action2LayoutSpriteType.BUILDING
            if name in ("xoffset", "yoffset", "zoffset"):
                if isinstance(value, expression.ConstantNumeric):
                    generic.check_range(value.value, -128, 127, name, value.pos)
                    return value
            else:
                assert name in ("xextent", "yextent", "zextent")
                if not isinstance(value, expression.ConstantNumeric):
                    raise generic.ScriptError(
                        "Value of '{}' must be a compile-time constant number.".format(name), value.pos
                    )
                generic.check_range(value.value, 0, 255, name, value.pos)
                return value
        # Value must be written to a register
        self.create_register(name, value)
        if self.type == Action2LayoutSpriteType.BUILDING:
            # For building sprites, x and y registers are always written together
            if name == "xoffset" and self.get_register("yoffset") is None:
                self.create_register("yoffset", expression.ConstantNumeric(0))
            if name == "yoffset" and self.get_register("xoffset") is None:
                self.create_register("xoffset", expression.ConstantNumeric(0))
        return expression.ConstantNumeric(0)

    def _validate_hide_sprite(self, name, value):
        self.create_register(name, expression.Not(value).reduce())


class ParsedSpriteLayout:
    def __init__(self):
        self.ground_sprite = None
        self.building_sprites = []
        self.advanced = False

    def get_size(self):
        size = 5
        if self.advanced:
            size += self.ground_sprite.get_registers_size()
        for sprite in self.building_sprites:
            if sprite.type == Action2LayoutSpriteType.CHILD:
                size += 7
            else:
                size += 10
            if self.advanced:
                size += sprite.get_registers_size()
        if len(self.building_sprites) == 0 and not self.advanced:
            size += 9
        return size

    def write(self, file):
        if self.advanced:
            file.print_byte(0x40 | len(self.building_sprites))
        else:
            file.print_byte(len(self.building_sprites))
        self.ground_sprite.write_sprite_number(file)
        if self.advanced:
            self.ground_sprite.write_flags(file)
            self.ground_sprite.write_registers(file)
        file.newline()
        if len(self.building_sprites) == 0 and not self.advanced:
            file.print_dwordx(0)  # sprite number 0 == no sprite
            for _ in range(0, 5):
                file.print_byte(0)  # empty bounding box. Note that number of zeros is 5, not 6
        else:
            for sprite in self.building_sprites:
                sprite.write_sprite_number(file)
                if self.advanced:
                    sprite.write_flags(file)
                file.print_byte(sprite.get_param("xoffset").value)
                file.print_byte(sprite.get_param("yoffset").value)
                if sprite.type == Action2LayoutSpriteType.CHILD:
                    file.print_bytex(0x80)
                else:
                    # normal building sprite
                    file.print_byte(sprite.get_param("zoffset").value)
                    file.print_byte(sprite.get_param("xextent").value)
                    file.print_byte(sprite.get_param("yextent").value)
                    file.print_byte(sprite.get_param("zextent").value)
                if self.advanced:
                    sprite.write_registers(file)
                file.newline()

    def process(self, spritelayout, feature, param_map, actions, layout_registers, var10map=None):
        if not isinstance(param_map, list):
            param_map = [param_map]

        # Reduce all expressions, can't do that earlier as feature is not known
        all_sprite_sets = []
        layout_sprite_list = []  # Create a new structure
        for layout_sprite in spritelayout.layout_sprite_list:
            param_list = []
            layout_sprite_list.append((layout_sprite.type, layout_sprite.pos, param_list))
            for param in layout_sprite.param_list:
                param_val = action2var.reduce_varaction2_expr(param.value, action2var.get_scope(feature), param_map)
                if isinstance(param_val, expression.SpriteGroupRef):
                    spriteset = action2.resolve_spritegroup(param_val.name)
                    if not spriteset.is_spriteset():
                        raise generic.ScriptError("Expected a reference to a spriteset.", param_val.pos)
                    if feature == 0x04:
                        assert var10map is not None
                        param_val = var10map.translate(spriteset, param_val.param_list, param_val.pos)
                    else:
                        all_sprite_sets.append(spriteset)
                param_list.append((param.name, param_val))

        actions.extend(action1.add_to_action1(all_sprite_sets, feature, spritelayout.pos))

        for type, pos, param_list in layout_sprite_list:
            if type.value not in layout_sprite_types:
                raise generic.ScriptError(
                    "Invalid sprite type '{}' encountered. Expected 'ground', 'building', or 'childsprite'.".format(
                        type.value
                    ),
                    type.pos,
                )
            sprite = Action2LayoutSprite(feature, layout_sprite_types[type.value], layout_registers, pos, param_map)
            for name, value in param_list:
                sprite.set_param(name, value)
            if sprite.type == Action2LayoutSpriteType.GROUND:
                if self.ground_sprite is not None:
                    raise generic.ScriptError("Sprite layout can have no more than one ground sprite", spritelayout.pos)
                self.ground_sprite = sprite
            else:
                self.building_sprites.append(sprite)

        if self.ground_sprite is None:
            if len(self.building_sprites) == 0:
                # no sprites defined at all, that's not very much.
                raise generic.ScriptError("Sprite layout requires at least one sprite", spritelayout.pos)
            # set to 0 for no ground sprite
            self.ground_sprite = Action2LayoutSprite(feature, Action2LayoutSpriteType.GROUND, layout_registers)
            self.ground_sprite.set_param(expression.Identifier("sprite"), expression.ConstantNumeric(0))

        self.advanced = any(x.is_advanced_sprite() for x in self.building_sprites + [self.ground_sprite])

    def write_action_value(self, actions, act6, offset):
        sprite_num = self.ground_sprite.get_sprite_number()
        sprite_num, offset = actionD.write_action_value(sprite_num, actions, act6, offset, 4)
        if self.advanced:
            offset += self.ground_sprite.get_registers_size()

        for sprite in self.building_sprites:
            sprite_num = sprite.get_sprite_number()
            sprite_num, offset = actionD.write_action_value(sprite_num, actions, act6, offset, 4)
            if self.advanced:
                offset += sprite.get_registers_size()
            offset += 3 if sprite.type == Action2LayoutSpriteType.CHILD else 6

        return offset


class LayoutRegisters:
    def __init__(self):
        self.registers = {}

    def add(self, value):
        if value not in self.registers:
            self.registers[value] = action2var.VarAction2StoreTempVar()
        return self.registers[value]

    def parse(self, varact2parser):
        for expr, reg in self.registers.items():
            varact2parser.parse_expr(expr)
            varact2parser.var_list.append(nmlop.STO_TMP)
            varact2parser.var_list.append(reg)
            varact2parser.var_list.append(nmlop.VAL2)
            varact2parser.var_list_size += reg.get_size() + 2


def get_layout_action2s(spritelayout, feature):
    actions = []

    if feature not in action2.features_sprite_layout:
        raise generic.ScriptError(
            "Sprite layouts are not supported for feature '{}'.".format(general.feature_name(feature))
        )

    # Allocate registers
    param_map = {}
    param_registers = []
    for param in spritelayout.param_list:
        reg = action2var.VarAction2CallParam(param.value)
        param_registers.append(reg)
        param_map[param.value] = reg
    param_map = (param_map, lambda name, value, pos: action2var.VarAction2LoadCallParam(value, name))
    spritelayout.register_map[feature] = param_registers

    layout_registers = LayoutRegisters()
    layout = ParsedSpriteLayout()
    layout.process(spritelayout, feature, param_map, actions, layout_registers)

    action6.free_parameters.save()
    act6 = action6.Action6()

    layout.write_action_value(actions, act6, 4)

    if len(act6.modifications) > 0:
        actions.append(act6)

    layout_action = Action2Layout(
        feature,
        spritelayout.name.value + " - feature {:02X}".format(feature),
        spritelayout.pos,
        layout,
        param_registers,
    )
    actions.append(layout_action)

    varact2parser = action2var.Varaction2Parser(feature)
    layout_registers.parse(varact2parser)

    # Only continue if we actually needed any new registers
    if varact2parser.var_list:
        # Remove the last VAL2 operator
        varact2parser.var_list.pop()
        varact2parser.var_list_size -= 1

        actions.extend(varact2parser.extra_actions)
        extra_act6 = action6.Action6()
        for mod in varact2parser.mods:
            extra_act6.modify_bytes(mod.param, mod.size, mod.offset + 4)
        if len(extra_act6.modifications) > 0:
            actions.append(extra_act6)

        varaction2 = action2var.Action2Var(
            feature, "{}@registers - feature {:02X}".format(spritelayout.name.value, feature), spritelayout.pos, 0x89
        )
        varaction2.var_list = varact2parser.var_list
        ref = expression.SpriteGroupRef(spritelayout.name, [], None, layout_action)
        varaction2.ranges.append(
            action2var.VarAction2Range(expression.ConstantNumeric(0), expression.ConstantNumeric(0), ref, "")
        )
        varaction2.default_result = ref
        varaction2.default_comment = ""

        # Add two references (default + range)
        # Make sure that registers allocated here are not used by the spritelayout
        action2.add_ref(ref, varaction2, True)
        action2.add_ref(ref, varaction2, True)
        spritelayout.set_action2(varaction2, feature)
        actions.append(varaction2)
    else:
        spritelayout.set_action2(layout_action, feature)

    action6.free_parameters.restore()
    return actions


def make_empty_layout_action2(feature, pos):
    """
    Make an empty layout action2
    For use with failed callbacks

    @param feature: Feature of the sprite layout to create
    @type feature: C{int}

    @param pos: Positional context.
    @type  pos: L{Position}

    @return: The created sprite layout action2
    @rtype: L{Action2Layout}
    """
    layout = ParsedSpriteLayout()
    layout.ground_sprite = Action2LayoutSprite(feature, Action2LayoutSpriteType.GROUND)
    layout.ground_sprite.set_param(expression.Identifier("sprite"), expression.ConstantNumeric(0))
    return Action2Layout(feature, "@CB_FAILED_LAYOUT{:02X}".format(feature), pos, layout, [])


class StationSpriteset(expression.Expression):
    def __init__(self, spriteset, args, var10, pos=None):
        expression.Expression.__init__(self, pos)
        self.spriteset = spriteset
        self.offset = args
        self.var10 = var10


class StationSpritesetVar10Map:
    def __init__(self):
        self.spritesets = {}
        self.var10 = 1  # Reserving 0 for SPRITESET() (basic action2)

    def translate(self, spriteset, args, pos):
        if spriteset not in self.spritesets:
            if self.var10 == 8:
                raise generic.ScriptError("A station can't use more than 6 different sprite sets", pos)
            self.spritesets[spriteset] = self.var10
            self.var10 += 1 if self.var10 != 1 else 2  # Reserving 2 for custom foundations
        return StationSpriteset(
            None if isinstance(spriteset, int) else spriteset, args, self.spritesets[spriteset], pos
        )

    def append_mapping(self, mapping, feature, actions, default, custom_spritesets):
        for spriteset, var10 in self.spritesets.items():
            if not isinstance(spriteset, int):
                if not spriteset.has_action2(feature):
                    actions.extend(action1.add_to_action1([spriteset], feature, None))
                    real_action2 = action2real.make_simple_real_action2(
                        feature,
                        spriteset.name.value + " - feature {:02X}".format(feature),
                        None,
                        action1.get_action1_index(spriteset, feature),
                    )
                    actions.append(real_action2)
                    spriteset.set_action2(real_action2, feature)
                ref = expression.SpriteGroupRef(spriteset.name, [], None, spriteset.get_action2(feature))
            else:
                if spriteset > len(custom_spritesets):
                    raise generic.ScriptError("Index out of range")
                ref = custom_spritesets[spriteset]
            # Skip default result
            if ref == default:
                continue
            mapping[var10] = (ref, None)
        return mapping


def parse_station_layouts(feature, id, layouts):
    var10map = StationSpritesetVar10Map()

    # Add DEFAULT([offset]) to reference active spriteset, selected by basic action2
    # and CUSTOM(index, [offset]) to reference a spriteset from the custom list
    def parse_spriteset(name, args, pos, info):
        if info:
            if len(args) < 1:
                raise generic.ScriptError("'{}' expects 1 or 2 parameters".format(name), pos)
            if not isinstance(args[0], expression.ConstantNumeric):
                raise generic.ScriptError("First parameter for '{}' must be a constant".format(name), pos)
            return var10map.translate(args[0].value, args[1:], pos)
        return StationSpriteset(None, args, info, pos)

    default_param = [
        (
            {"DEFAULT": None, "CUSTOM": True},
            lambda name, value, pos: expression.FunctionPtr(expression.Identifier(name, pos), parse_spriteset, value),
        )
    ]

    actions = []
    param_registers = []
    parsed_layouts = []
    varact2parser = action2var.Varaction2Parser(feature)
    layout_registers = LayoutRegisters()

    for layout in layouts:
        spritelayout = (
            action2.resolve_spritegroup(layout.name) if isinstance(layout, expression.SpriteGroupRef) else None
        )
        if not spritelayout or not isinstance(spritelayout, spriteblock.SpriteLayout):
            raise generic.ScriptError("Expected a SpriteLayout", layout.pos)

        # Allocate registers
        param_map = {}
        for i, param in enumerate(spritelayout.param_list):
            reg = action2var.VarAction2CallParam(param.value)
            param_registers.append(reg)
            param_map[param.value] = reg
            store_tmp = action2var.VarAction2StoreCallParam(reg)
            varact2parser.parse_expr(layout.param_list[i])
            varact2parser.var_list.append(nmlop.STO_TMP)
            varact2parser.var_list.append(store_tmp)
            varact2parser.var_list.append(nmlop.VAL2)
            varact2parser.var_list_size += store_tmp.get_size() + 2
        param_map = [(param_map, lambda name, value, pos: action2var.VarAction2LoadCallParam(value, name))]
        param_map.extend(default_param)

        layout = ParsedSpriteLayout()
        layout.process(spritelayout, feature, param_map, actions, layout_registers, var10map)
        parsed_layouts.append(layout)

    layout_registers.parse(varact2parser)

    actions.extend(action0.get_layout_action0(feature, id, parsed_layouts))

    registers_ref = None

    # Create a procedure only if needed
    if varact2parser.var_list:
        # Remove the last VAL2 operator
        varact2parser.var_list.pop()
        varact2parser.var_list_size -= 1

        actions.extend(varact2parser.extra_actions)
        extra_act6 = action6.Action6()
        for mod in varact2parser.mods:
            extra_act6.modify_bytes(mod.param, mod.size, mod.offset + 4)
        if len(extra_act6.modifications) > 0:
            actions.append(extra_act6)

        varaction2 = action2var.Action2Var(
            feature, "Station Layout@registers - Id {:02X}".format(id.value), None, 0x89, param_registers
        )
        varaction2.var_list = varact2parser.var_list
        varaction2.default_result = expression.ConstantNumeric(0)
        varaction2.default_comment = "Return computed value"

        actions.append(varaction2)
        registers_ref = expression.SpriteGroupRef(expression.Identifier(varaction2.name), [], None, varaction2)

    return (actions, var10map, registers_ref)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic
from nml.actions import action12
from nml.ast import base_statement, sprite_container


class FontGlyphBlock(base_statement.BaseStatement, sprite_container.SpriteContainer):
    """
    AST class for a font_glyph block
    Syntax: font_glyph(

    @ivar font_size: Size of the font to provide characters for (NORMAL/SMALL/LARGE/MONO)
    @type font_size: L{Expression}

    @ivar base_char: First character to replace
    @type base_char: L{Expression}

    @ivar image_file: Default file to use for the contained real sprites (none if N/A)
    @type image_file: L{StringLiteral} or C{None}

    @ivar sprite_list: List of real sprites
    @type sprite_list: C{list} of L{RealSprite}
    """

    def __init__(self, param_list, sprite_list, name, pos):
        base_statement.BaseStatement.__init__(self, "font_glyph-block", pos)
        sprite_container.SpriteContainer.__init__(self, "font_glyph-block", name)
        if not (2 <= len(param_list) <= 3):
            raise generic.ScriptError(
                "font_glyph-block requires 2 or 3 parameters, encountered " + str(len(param_list)), pos
            )
        self.font_size = param_list[0]
        self.base_char = param_list[1]
        self.image_file = param_list[2].reduce() if len(param_list) >= 3 else None
        if self.image_file is not None and not isinstance(self.image_file, expression.StringLiteral):
            raise generic.ScriptError(
                "font_glyph-block parameter 3 'file' must be a string literal", self.image_file.pos
            )
        self.sprite_list = sprite_list
        self.add_sprite_data(self.sprite_list, self.image_file, pos)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Load font glyphs, starting at", self.base_char)
        generic.print_dbg(indentation + 2, "Font size:  ", self.font_size)
        generic.print_dbg(
            indentation + 2, "Source:  ", self.image_file.value if self.image_file is not None else "None"
        )

        generic.print_dbg(indentation + 2, "Sprites:")
        for sprite in self.sprite_list:
            sprite.debug_print(indentation + 4)

    def get_action_list(self):
        return action12.parse_action12(self)

    def __str__(self):
        name = str(self.block_name) if self.block_name is not None else ""
        params = [self.font_size, self.base_char]
        if self.image_file is not None:
            params.append(self.image_file)
        ret = "font_glyph {}({}) {{\n".format(name, ", ".join(str(param) for param in params))
        for sprite in self.sprite_list:
            ret += "\t{}\n".format(sprite)
        ret += "}\n"
        return ret



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import output_base


class OutputNML(output_base.TextOutputBase):
    """
    Class for outputting NML.
    """

    def __init__(self, filename):
        output_base.TextOutputBase.__init__(self, filename)

    def open_file(self):
        return open(self.filename, "w", encoding="utf-8")

    def write(self, text):
        self.file.write(text)

    def newline(self):
        self.file.write("\n")



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, grfstrings
from nml.actions import base_action


class Action14(base_action.BaseAction):
    def __init__(self, nodes):
        self.nodes = nodes

    def skip_action7(self):
        return False

    def write(self, file):
        size = 2  # final 0-byte
        for node in self.nodes:
            size += node.get_size()

        file.start_sprite(size)
        file.print_bytex(0x14)
        for node in self.nodes:
            node.write(file)
        file.print_bytex(0)

        file.end_sprite()


def split_action14(node, max_size):
    if node.get_size() <= max_size:
        return [node, None]
    if not isinstance(node, BranchNode):
        return [None, node]
    new_node = BranchNode(node.id)
    rest = BranchNode(node.id)
    copy_to_rest = False
    for subnode in node.subnodes:
        if copy_to_rest:
            rest.subnodes.append(subnode)
            continue

        new_subnode, subnode_rest = split_action14(subnode, max_size - new_node.get_size())
        if new_subnode is not None:
            new_node.subnodes.append(new_subnode)
        if subnode_rest is not None:
            rest.subnodes.append(subnode_rest)
            copy_to_rest = True

    assert len(rest.subnodes) > 0
    if len(new_node.subnodes) == 0:
        return [None, rest]
    return [new_node, rest]


def get_actions(root):
    action_list = []
    while True:
        node, root = split_action14(root, 65535)
        assert node is not None
        action_list.append(Action14([node]))
        if root is None:
            break

    return action_list


class Action14Node:
    def __init__(self, type_string, id):
        self.type_string = type_string
        self.id = id

    def get_size(self):
        """
        How many bytes will be written to the output file by L{write}?

        @return: The size (in bytes) of this node.
        """
        raise NotImplementedError("get_size must be implemented in Action14Node-subclass {!r}".format(type(self)))

    def write(self, file):
        """
        Write this node to the output file.

        @param file: The file to write the output to.
        """
        raise NotImplementedError("write must be implemented in Action14Node-subclass {!r}".format(type(self)))

    def write_type_id(self, file):
        file.print_string(self.type_string, False, True)
        if isinstance(self.id, str):
            file.print_string(self.id, False, True)
        else:
            file.print_dword(self.id)


class TextNode(Action14Node):
    def __init__(self, id, string, skip_default_langid=False):
        Action14Node.__init__(self, "T", id)
        self.string = string
        grfstrings.validate_string(self.string)
        self.skip_default_langid = skip_default_langid

    def get_size(self):
        if self.skip_default_langid:
            size = 0
        else:
            size = 6 + grfstrings.get_string_size(grfstrings.get_translation(self.string))
        for lang_id in grfstrings.get_translations(self.string):
            # 6 is for "T" (1), id (4), langid (1)
            size += 6 + grfstrings.get_string_size(grfstrings.get_translation(self.string, lang_id))
        return size

    def write(self, file):
        if not self.skip_default_langid:
            self.write_type_id(file)
            file.print_bytex(0x7F)
            file.print_string(grfstrings.get_translation(self.string))
            file.newline()
        for lang_id in grfstrings.get_translations(self.string):
            self.write_type_id(file)
            file.print_bytex(lang_id)
            file.print_string(grfstrings.get_translation(self.string, lang_id))
            file.newline()


class BranchNode(Action14Node):
    def __init__(self, id):
        Action14Node.__init__(self, "C", id)
        self.subnodes = []

    def get_size(self):
        size = 6  # "C", id, final 0-byte
        for node in self.subnodes:
            size += node.get_size()
        return size

    def write(self, file):
        self.write_type_id(file)
        file.newline()
        for node in self.subnodes:
            node.write(file)
        file.print_bytex(0)
        file.newline()


class BinaryNode(Action14Node):
    def __init__(self, id, size, val=None):
        Action14Node.__init__(self, "B", id)
        self.size = size
        self.val = val

    def get_size(self):
        return 7 + self.size  # "B" (1), id (4), size (2), data (self.size)

    def write(self, file):
        self.write_type_id(file)
        file.print_word(self.size)
        file.print_varx(self.val, self.size)
        file.newline()


class UsedPaletteNode(BinaryNode):
    def __init__(self, pal):
        BinaryNode.__init__(self, "PALS", 1)
        self.pal = pal

    def write(self, file):
        self.write_type_id(file)
        file.print_word(self.size)
        file.print_string(self.pal, False, True)
        file.newline()


class BlitterNode(BinaryNode):
    def __init__(self, blitter):
        BinaryNode.__init__(self, "BLTR", 1)
        self.blitter = blitter

    def write(self, file):
        self.write_type_id(file)
        file.print_word(self.size)
        file.print_string(self.blitter, False, True)
        file.newline()


class SettingMaskNode(BinaryNode):
    def __init__(self, param_num, first_bit, num_bits):
        BinaryNode.__init__(self, "MASK", 3)
        self.param_num = param_num
        self.first_bit = first_bit
        self.num_bits = num_bits

    def write(self, file):
        self.write_type_id(file)
        file.print_word(self.size)
        file.print_byte(self.param_num)
        file.print_byte(self.first_bit)
        file.print_byte(self.num_bits)
        file.newline()


class LimitNode(BinaryNode):
    def __init__(self, min_val, max_val):
        BinaryNode.__init__(self, "LIMI", 8)
        self.min_val = min_val
        self.max_val = max_val

    def write(self, file):
        self.write_type_id(file)
        file.print_word(self.size)
        file.print_dword(self.min_val)
        file.print_dword(self.max_val)
        file.newline()


def grf_name_desc_actions(root, name, desc, url, version, min_compatible_version):
    if len(grfstrings.get_translations(name)) > 0:
        name_node = TextNode("NAME", name, True)
        root.subnodes.append(name_node)
    if len(grfstrings.get_translations(desc)) > 0:
        desc_node = TextNode("DESC", desc, True)
        root.subnodes.append(desc_node)
    if url is not None:
        desc_node = TextNode("URL_", url)
        root.subnodes.append(desc_node)
    version_node = BinaryNode("VRSN", 4, version.value)
    root.subnodes.append(version_node)
    min_compatible_version_node = BinaryNode("MINV", 4, min_compatible_version.value)
    root.subnodes.append(min_compatible_version_node)


def param_desc_actions(root, params):
    num_params = 0
    for param_desc in params:
        num_params += len(param_desc.setting_list)
    root.subnodes.append(BinaryNode("NPAR", 1, num_params))
    param_root = BranchNode("PARA")
    param_num = 0
    setting_num = 0
    for param_desc in params:
        if param_desc.num is not None:
            param_num = param_desc.num.value
        for setting in param_desc.setting_list:
            setting_node = BranchNode(setting_num)
            if setting.name_string is not None:
                setting_node.subnodes.append(TextNode("NAME", setting.name_string))
            if setting.desc_string is not None:
                setting_node.subnodes.append(TextNode("DESC", setting.desc_string))
            if setting.type == "int":
                setting_node.subnodes.append(BinaryNode("MASK", 1, param_num))
                min_val = setting.min_val.uvalue if setting.min_val is not None else 0
                max_val = setting.max_val.uvalue if setting.max_val is not None else 0xFFFFFFFF
                def_val = setting.def_val.uvalue if setting.def_val is not None else 0
                if min_val > max_val or def_val < min_val or def_val > max_val:
                    generic.print_warning(
                        generic.Warning.GENERIC,
                        "Limits for GRF parameter {} are incoherent, ignoring.".format(param_num),
                    )
                    min_val = 0
                    max_val = 0xFFFFFFFF
                setting_node.subnodes.append(LimitNode(min_val, max_val))
                if len(setting.val_names) > 0:
                    value_names_node = BranchNode("VALU")
                    for set_val_pair in setting.val_names:
                        value_names_node.subnodes.append(TextNode(set_val_pair[0], set_val_pair[1]))
                    setting_node.subnodes.append(value_names_node)
            else:
                assert setting.type == "bool"
                assert setting.bit_num is not None
                setting_node.subnodes.append(BinaryNode("TYPE", 1, 1))
                setting_node.subnodes.append(SettingMaskNode(param_num, setting.bit_num.value, 1))
            if setting.def_val is not None:
                setting_node.subnodes.append(BinaryNode("DFLT", 4, setting.def_val.value))
            param_root.subnodes.append(setting_node)
            setting_num += 1
        param_num += 1
    if len(param_root.subnodes) > 0:
        root.subnodes.append(param_root)


def PaletteAction(pal):
    root = BranchNode("INFO")
    pal_node = UsedPaletteNode(pal)
    root.subnodes.append(pal_node)
    return [Action14([root])]



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

"""
Code for storing and generating action F
"""
from nml import expression, generic, grfstrings
from nml.actions import base_action

# Helper functions to allocate townname IDs
#
# Numbers available are from 0 to 0x7f (inclusive).
# These numbers can be in five states:
# - free:            Number is available for use.
# - named:           Number is allocated to represent a name.
# - safe numbered:   Number is allocated by the user, and is safe to refer to.
# - unsafe numbered: Number is allocated by the user, and is not safe to refer to
#                    (that is, it is below the point of 'prepare_output')
# - invisible:       Number is allocated by a final town_name, without attaching a name to it.
#                    It is not accessible anymore.
# Instances of the TownNames class have a 'name' attribute, which can be 'None' (for an invisible number),
# a string (for a named number), or a (constant numeric) expression (for a safe/unsafe number).
#
total_numbers = 0x80
free_numbers = set(range(total_numbers))  #: Free numbers.
first_free_id = 0  #: All numbers before this are allocated.
named_numbers = {}  #: Mapping of names to named numbers. Note that these are always safe to refer to.
numbered_numbers = set()  #: Safe numbers introduced by the user (without name).


def get_free_id():
    """Allocate a number from the free_numbers."""
    global first_free_id
    while first_free_id not in free_numbers:
        first_free_id = first_free_id + 1
        if first_free_id >= total_numbers:
            raise generic.ScriptError(
                "Too many town name blocks. Some of these are autogenerated. Please use less town names."
            )
    number = first_free_id
    free_numbers.remove(number)
    first_free_id = first_free_id + 1
    return number


town_names_blocks = {}  # Mapping of town_names ID number to TownNames instance.


def print_stats():
    """
    Print statistics about used ids.
    """
    num_used = total_numbers - len(free_numbers)
    if num_used > 0:
        generic.print_info("Town names: {}/{}".format(num_used, total_numbers))


class ActionF(base_action.BaseAction):
    """
    Town names action.

    @ivar name: Name ID of the town_name.
    @type name: C{None}, L{Identifier}, or L{ConstantNumeric}

    @ivar id_number: Allocated ID number for this town_name action F node.
    @type id_number: C{None} or C{int}

    @ivar style_name: Name of the translated string containing the name of the style, if any.
    @type style_name: C{None} or L{String}

    @ivar style_names: List translations of L{style_name}, pairs (languageID, text).
    @type style_names: C{list} of (C{int}, L{Identifier})

    @ivar parts: Parts of the names.
    @type parts: C{list} of L{TownNamesPart}

    @ivar pos: Position information of the 'town_names' block.
    @type pos: L{Position}

    @ivar deps: List of ID numbers of action F dependant on this node.
    @type deps: C{list} of C{int}
    """

    def __init__(self, name, id_number, style_name, parts, pos):
        self.name = name
        self.id_number = id_number
        self.style_name = style_name
        self.style_names = []
        self.parts = parts
        self.pos = pos
        self.deps = []

    def prepare_output(self, sprite_num):
        # Resolve references to earlier townname actions
        referenced_blocks = []
        for part in self.parts:
            part_blocks = part.resolve_townname_id()
            if part_blocks:
                referenced_blocks.append(part_blocks)

        # Allocate a number for this action F.
        if self.name is None or isinstance(self.name, expression.Identifier):
            self.id_number = get_free_id()
            if isinstance(self.name, expression.Identifier):
                if self.name.value in named_numbers:
                    raise generic.ScriptError(
                        'Cannot define town name "{}", it is already in use'.format(self.name), self.pos
                    )
                named_numbers[self.name.value] = self.id_number  # Add name to the set 'safe' names.
        else:
            numbered_numbers.add(self.id_number)  # Add number to the set of 'safe' numbers.

        town_names_blocks[self.id_number] = self  # Add self to the available blocks.

        def assign_bits(part, startbit):
            num_bits = part.assign_bits(startbit)
            # Prevent overlap if really needed
            if len(part.pieces) > 1:
                startbit += num_bits
            return startbit

        def update_block(block, startbit):
            for part in block.parts:
                startbit = assign_bits(part, startbit)
                # Update dependant Action F
                for dep in block.deps:
                    update_block(town_names_blocks[dep], startbit)
            return startbit

        # Determine lowest available bit, and prevent overlap when needed.
        # referenced_blocks is a list of sets, each set contains references that can share the same bits,
        # but different sets can't share bits.
        startbit = 0
        for part_blocks in referenced_blocks:
            freebit = startbit
            for part_block in part_blocks:
                block = town_names_blocks[part_block]
                startbit = max(startbit, update_block(block, freebit))
                if self.id_number not in block.deps:
                    block.deps.append(self.id_number)

        # Allocate random bits to all parts.
        for part in self.parts:
            startbit = assign_bits(part, startbit)

        if startbit > 32:
            raise generic.ScriptError(
                "Not enough random bits for the town name generation ({:d} needed, 32 available)".format(startbit),
                self.pos,
            )

        # Pull style names if needed.
        if self.style_name is not None:
            grfstrings.validate_string(self.style_name)
            self.style_names = [
                (lang_id, grfstrings.get_translation(self.style_name, lang_id))
                for lang_id in grfstrings.get_translations(self.style_name)
            ]
            self.style_names.append((0x7F, grfstrings.get_translation(self.style_name)))
            self.style_names.sort()
            if len(self.style_names) == 0:
                raise generic.ScriptError(
                    'Style "{}" defined, but no translations found for it'.format(self.style_name.name.value), self.pos
                )
        else:
            self.style_names = []

    # Style names
    def get_length_styles(self):
        if len(self.style_names) == 0:
            return 0

        size = 0
        for _lang, txt in self.style_names:
            size += 1 + grfstrings.get_string_size(txt)  # Language ID, text
        return size + 1  # Terminating 0

    def write_styles(self, handle):
        if len(self.style_names) == 0:
            return

        for lang, txt in self.style_names:
            handle.print_bytex(lang)
            handle.print_string(txt, final_zero=True)
            handle.newline()
        handle.print_bytex(0)
        handle.newline()

    # Parts
    def get_length_parts(self):
        size = 1  # num_parts byte
        return size + sum(part.get_length() for part in self.parts)

    def write_parts(self, handle):
        handle.print_bytex(len(self.parts))
        for part in self.parts:
            part.write(handle)
            handle.newline()

    def write(self, handle):
        handle.start_sprite(2 + self.get_length_styles() + self.get_length_parts())
        handle.print_bytex(0xF)
        handle.print_bytex(self.id_number | (0x80 if len(self.style_names) > 0 else 0))
        handle.newline(str(self.name) if self.name is not None else "")
        self.write_styles(handle)
        self.write_parts(handle)
        handle.end_sprite()

    def skip_action7(self):
        return False



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants, nmlop
from nml.actions import action2, action2var, action6, actionD


class Action2Production(action2.Action2):
    """
    Class corresponding to Action2Industries (=production CB)

    @ivar version: Production CB version. Version 0 uses constants, version 1 uses registers.
    @type version: C{int}

    @ivar sub_in: Amounts (v0) or registers (v1) to subtract from incoming cargos.
    @type sub_in: C{list} of (C{int} or L{VarAction2Var})

    @ivar add_out: Amounts (v0) or registers (v1) to add to the output cargos.
    @type add_out: C{list} of (C{int} or L{VarAction2Var})

    @ivar again: Number (v0) or register (v1), production CB will be run again if nonzero.
    @type again C{int} or L{VarAction2Var}
    """

    def __init__(self, name, pos, version, sub_in, add_out, again):
        action2.Action2.__init__(self, 0x0A, name, pos)
        self.version = version
        if version in (0, 1):
            self.sub_in = sub_in
            assert len(self.sub_in) == 3
            self.add_out = add_out
            assert len(self.add_out) == 2
        elif version == 2:
            self.sub_in = sub_in
            self.add_out = add_out
        else:
            raise AssertionError()
        self.again = again

    def prepare_output(self, sprite_num):
        action2.Action2.prepare_output(self, sprite_num)

    def write(self, file):
        if self.version in (0, 1):
            cargo_size = 2 if self.version == 0 else 1
            size = 2 + 5 * cargo_size
            action2.Action2.write_sprite_start(self, file, size)
            file.print_bytex(self.version)
            values = self.sub_in + self.add_out + [self.again]
            # Read register numbers if needed
            if self.version == 1:
                values = [val.parameter for val in values]

            for val in values[:-1]:
                file.print_varx(val, cargo_size)
            file.print_bytex(values[-1])
        elif self.version == 2:
            size = 4 + 2 * (len(self.sub_in) + len(self.add_out))
            action2.Action2.write_sprite_start(self, file, size)
            file.print_bytex(self.version)
            file.print_byte(len(self.sub_in))
            for cargoindex, value in self.sub_in:
                file.print_bytex(cargoindex)
                file.print_bytex(value.parameter)
            file.print_byte(len(self.add_out))
            for cargoindex, value in self.add_out:
                file.print_bytex(cargoindex)
                file.print_bytex(value.parameter)
            file.print_bytex(self.again.parameter)
        file.newline()
        file.end_sprite()


def resolve_prodcb_register(param, varact2parser):
    if (
        isinstance(param, expression.StorageOp)
        and param.name == "LOAD_TEMP"
        and isinstance(param.register, expression.ConstantNumeric)
    ):
        # We can load a register directly
        res = action2var.VarAction2Var(0x7D, 0, 0xFFFFFFFF, param.register.value)
    else:
        if len(varact2parser.var_list) != 0:
            varact2parser.var_list.append(nmlop.VAL2)
            varact2parser.var_list_size += 1
        varact2parser.parse_expr(action2var.reduce_varaction2_expr(param, action2var.get_scope(0x0A)))
        store_tmp = action2var.VarAction2StoreTempVar()
        res = action2var.VarAction2LoadTempVar(store_tmp)
        varact2parser.var_list.append(nmlop.STO_TMP)
        varact2parser.var_list.append(store_tmp)
        varact2parser.var_list_size += store_tmp.get_size() + 1  # Add 1 for operator
    return res


def finish_production_actions(produce, prod_action, action_list, varact2parser):
    action_list.append(prod_action)

    if len(varact2parser.var_list) == 0:
        produce.set_action2(prod_action, 0x0A)
    else:
        # Create intermediate varaction2
        varaction2 = action2var.Action2Var(0x0A, "{}@registers".format(produce.name.value), produce.pos, 0x89)
        varaction2.var_list = varact2parser.var_list
        action_list.extend(varact2parser.extra_actions)
        extra_act6 = action6.Action6()
        for mod in varact2parser.mods:
            extra_act6.modify_bytes(mod.param, mod.size, mod.offset + 4)
        if len(extra_act6.modifications) > 0:
            action_list.append(extra_act6)
        ref = expression.SpriteGroupRef(produce.name, [], None, prod_action)
        varaction2.ranges.append(
            action2var.VarAction2Range(expression.ConstantNumeric(0), expression.ConstantNumeric(0), ref, "")
        )
        varaction2.default_result = ref
        varaction2.default_comment = ""

        # Add two references (default + range)
        action2.add_ref(ref, varaction2)
        action2.add_ref(ref, varaction2)
        produce.set_action2(varaction2, 0x0A)
        action_list.append(varaction2)

    action6.free_parameters.restore()

    return action_list


def get_production_actions(produce):
    """
    Get the action list that implements the given produce-block in nfo.

    @param produce: Produce-block to parse.
    @type produce: L{Produce}
    """
    action_list = []
    act6 = action6.Action6()
    action6.free_parameters.save()

    result_list = []
    varact2parser = action2var.Varaction2Parser(0x0A)
    if all(x.supported_by_actionD(False) for x in produce.param_list):
        version = 0
        offset = 4
        for i, param in enumerate(produce.param_list):
            result, offset = actionD.write_action_value(param, action_list, act6, offset, 2 if i < 5 else 1)
            result_list.append(result.value)
    else:
        version = 1
        for param in produce.param_list:
            result_list.append(resolve_prodcb_register(param, varact2parser))

    if len(act6.modifications) > 0:
        action_list.append(act6)
    prod_action = Action2Production(
        produce.name.value, produce.pos, version, result_list[0:3], result_list[3:5], result_list[5]
    )

    return finish_production_actions(produce, prod_action, action_list, varact2parser)


def get_production_v2_actions(produce):
    """
    Get the action list that implements the given produce-block in nfo.

    @param produce: Produce-block to parse.
    @type produce: L{Produce2}
    """
    action_list = []
    action6.free_parameters.save()

    varact2parser = action2var.Varaction2Parser(0x0A)

    def resolve_cargoitem(item):
        cargolabel = item.name.value
        if cargolabel not in global_constants.cargo_numbers:
            raise generic.ScriptError("Cargo label {0} not found in your cargo table".format(cargolabel), produce.pos)
        cargoindex = global_constants.cargo_numbers[cargolabel]
        valueregister = resolve_prodcb_register(item.value, varact2parser)
        return (cargoindex, valueregister)

    sub_in = [resolve_cargoitem(item) for item in produce.subtract_in]
    add_out = [resolve_cargoitem(item) for item in produce.add_out]
    again = resolve_prodcb_register(produce.again, varact2parser)

    prod_action = Action2Production(produce.name.value, produce.pos, 2, sub_in, add_out, again)

    return finish_production_actions(produce, prod_action, action_list, varact2parser)


def make_empty_production_action2(pos):
    """
    Make an empty production action2
    For use with failed callbacks

    @param pos: Positional context.
    @type  pos: L{Position}

    @return: The created production action2
    @rtype: L{Action2Production}
    """
    return Action2Production("@CB_FAILED_PROD", pos, 0, [0, 0, 0], [0, 0], 0)
